#define _TEST_E_FONT_C_

#include <gdk/gdkx.h>
#include <gnome.h>
#include "e-font.h"

static void print_gdk_font_name (const GdkFont * font, const gchar * text);

int main (int argc, char ** argv)
{
	char **namelist;
	int numfonts;
	int i;

	gdk_init (&argc, &argv);

	namelist = XListFonts (GDK_DISPLAY (),
			       "*",
			       16384, &numfonts);

	g_print ("Loaded %d font names\n", numfonts);

	for (i = 0; i < numfonts; i++) {
		GdkFont * gdkfont;
		iconv_t iconv;
		gdkfont = gdk_font_load (namelist[i]);
		iconv = e_iconv_from_gdk_font (gdkfont);
		if (iconv == (iconv_t) -1) {
			print_gdk_font_name (gdkfont, "Missing from -");
		}
		iconv = e_iconv_to_gdk_font (gdkfont);
		if (iconv == (iconv_t) -1) {
			print_gdk_font_name (gdkfont, "Missing to -");
		}
		gdk_font_unref (gdkfont);
	}

	XFreeFontNames (namelist);

	return 0;
}

static void
print_gdk_font_name (const GdkFont * font, const gchar * text)
{
	Atom font_atom, atom;
	Bool status;

	font_atom = gdk_atom_intern ("FONT", FALSE);

	g_print (text);

	if (font == NULL) {
		g_print (" font is NULL\n");
	} else if (font->type == GDK_FONT_FONTSET) {
		XFontStruct **font_structs;
		gint num_fonts;
		gchar **font_names;

		num_fonts = XFontsOfFontSet (GDK_FONT_XFONT (font), &font_structs, &font_names);

		g_print (" fontset: %s\n",font_names[0]);
	} else {
		gchar * name;
		status = XGetFontProperty (GDK_FONT_XFONT (font), font_atom, &atom);
		name = gdk_atom_name (atom);
		g_print (" font: %s\n", name);
		if (name) g_free (name);
	}
}
